@@ -5,60 +5,44 @@
  * Copyright (c) 2007-2008 Neil Conway <neilc@samurai.com>
  * Copyright (c) 2007 Open Technology Group, Inc. <http://www.otg-nc.com>
  * Copyright (c) 2008-2013 Hannu Valtonen <hannu.valtonen@ohmu.fi>
+ * Copyright (c) 2012-2014 Ohmu Ltd <opensource@ohmu.fi>
  *
  * See the file LICENSE for distribution terms.
  */
+
 #include "pgmemcache.h"
 
 #ifdef PG_MODULE_MAGIC
 PG_MODULE_MAGIC;
 #endif
 
+/* Internal functions */
+static void pgmemcache_reset_context(void);
+static void assign_sasl_params(const char *username, const char *password);
+static void assign_default_servers_guc(const char *newval, void *extra);
+static void assign_default_behavior_guc(const char *newval, void *extra);
+static memcached_behavior get_memcached_behavior_flag(const char *flag);
+static uint64_t get_memcached_behavior_data(const char *flag, const char *data);
+static uint64_t get_memcached_hash_type(const char *data);
+static uint64_t get_memcached_distribution_type(const char *data);
+static Datum memcache_atomic_op(bool increment, PG_FUNCTION_ARGS);
+static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS);
+static memcached_return do_server_add(const char *host_str);
+
 /* Per-backend global state. */
 static struct memcache_global_s
 {
   memcached_st *mc;
-  /* context in which long-lived state is allocated */
-  MemoryContext pg_ctxt;
   char *default_servers;
   char *default_behavior;
   char *sasl_authentication_username;
   char *sasl_authentication_password;
 } globals;
 
+
 void _PG_init(void)
 {
-  MemoryContext old_ctxt;
-  int rc;
-
-  globals.pg_ctxt = AllocSetContextCreate(TopMemoryContext,
-                                          "pgmemcache global context",
-                                          ALLOCSET_SMALL_MINSIZE,
-                                          ALLOCSET_SMALL_INITSIZE,
-                                          ALLOCSET_SMALL_MAXSIZE);
-
-  old_ctxt = MemoryContextSwitchTo(globals.pg_ctxt);
-  globals.mc = memcached_create(NULL);
-
-  if (memcached_set_memory_allocators(globals.mc,
-                                      (memcached_malloc_fn) pgmemcache_malloc,
-                                      (memcached_free_fn) pgmemcache_free,
-                                      (memcached_realloc_fn) pgmemcache_realloc,
-                                      (memcached_calloc_fn) pgmemcache_calloc,
-                                      NULL) != MEMCACHED_SUCCESS) {
-    elog(ERROR, "pgmemcache: unable to set memory allocators");
-  }
-
-  MemoryContextSwitchTo(old_ctxt);
-
-  /* Use memcache binary protocol by default as required for
-     memcached_(increment|decrement)_with_initial. */
-  rc = memcached_behavior_set(globals.mc,
-                              MEMCACHED_BEHAVIOR_BINARY_PROTOCOL,
-                              1);
-  if (rc != MEMCACHED_SUCCESS)
-    elog(WARNING, "pgmemcache: memcached_behavior_set(BINARY_PROTOCOL): %s",
-                  memcached_strerror(globals.mc, rc));
+  pgmemcache_reset_context();
 
   DefineCustomStringVariable("pgmemcache.default_servers",
                              "Comma-separated list of memcached servers to connect to.",
@@ -112,15 +96,72 @@ void _PG_init(void)
                              NULL,
                              NULL);
 
-#if LIBMEMCACHED_WITH_SASL_SUPPORT
-  /* XXX: does this work?  We should probably add an assign_hook for these
-   * so that memcache sasl data is updated when the values are changed. */
-  if (globals.sasl_authentication_username != NULL && strlen(globals.sasl_authentication_username) > 0 &&
-      globals.sasl_authentication_password != NULL && strlen(globals.sasl_authentication_password) > 0)
+  /* XXX: We should have real assign_hook for these so that memcache sasl data
+   * is updated when the values are changed.  */
+  assign_sasl_params(globals.sasl_authentication_username, globals.sasl_authentication_password);
+}
+
+/* This is called when we're being unloaded from a process. Note that
+ * this only happens when we're being replaced by a LOAD (e.g. it
+ * doesn't happen on process exit), so we can't depend on it being
+ * called. */
+void _PG_fini(void)
+{
+  memcached_free(globals.mc);
+}
+
+static time_t interval_to_time_t(Interval *span)
+{
+  float8 result;
+
+#ifdef HAVE_INT64_TIMESTAMP
+  result = span->time / 1000000e0;
+#else
+  result = span->time;
+#endif
+
+  result += span->day * 86400;
+
+  if (span->month != 0)
     {
-      int rc = memcached_set_sasl_auth_data(globals.mc,
-                                            globals.sasl_authentication_username,
-                                            globals.sasl_authentication_password);
+      result += (365.25 * 86400) * (span->month / 12);
+      result += (30.0 * 86400) * (span->month % 12);
+    }
+
+  return (time_t) result;
+}
+
+static void pgmemcache_reset_context(void)
+{
+  int rc;
+
+  if (globals.mc)
+    {
+      memcached_free(globals.mc);
+      globals.mc = NULL;
+    }
+
+  globals.mc = memcached_create(NULL);
+
+  /* Always use the memcache binary protocol as required for
+     memcached_(increment|decrement)_with_initial. */
+  rc = memcached_behavior_set(globals.mc,
+                              MEMCACHED_BEHAVIOR_BINARY_PROTOCOL,
+                              1);
+  if (rc != MEMCACHED_SUCCESS)
+    elog(WARNING, "pgmemcache: memcached_behavior_set(BINARY_PROTOCOL): %s",
+                  memcached_strerror(globals.mc, rc));
+
+  assign_default_behavior_guc(globals.default_behavior, NULL);
+  assign_sasl_params(globals.sasl_authentication_username, globals.sasl_authentication_password);
+}
+
+static void assign_sasl_params(const char *username, const char *password)
+{
+#if LIBMEMCACHED_WITH_SASL_SUPPORT
+  if (username != NULL && strlen(username) > 0 && password != NULL && strlen(password) > 0)
+    {
+      int rc = memcached_set_sasl_auth_data(globals.mc, username, password);
       if (rc != MEMCACHED_SUCCESS)
         elog(ERROR, "pgmemcache: memcached_set_sasl_auth_data: %s",
                     memcached_strerror(globals.mc, rc));
@@ -131,31 +172,15 @@ void _PG_init(void)
 #endif
 }
 
-static void *pgmemcache_malloc(memcached_st *ptr __attribute__((unused)), const size_t size, void *context)
-{
-  return MemoryContextAllocZero(globals.pg_ctxt, size);
-}
-
-static void pgmemcache_free(memcached_st *ptr __attribute__((unused)), void *mem, void *context)
-{
-  pfree(mem);
-}
-
-static void *pgmemcache_realloc(memcached_st *ptr __attribute__((unused)), void *mem, const size_t size, void *context)
-{
-  /* postgresql repalloc() fails if 'mem' is NULL */
-  return mem ? repalloc(mem, (Size)size) : MemoryContextAllocZero(globals.pg_ctxt, size);
-}
-
-static void *pgmemcache_calloc(memcached_st *ptr __attribute__((unused)), size_t nelem, const size_t size, void *context)
-{
-  return MemoryContextAllocZero(globals.pg_ctxt, nelem * size);
-}
-
 static void assign_default_servers_guc(const char *newval, void *extra)
 {
   if (newval)
-    do_server_add(newval);
+    {
+      /* there is no way to remove servers from a memcache context, so
+       * recreate it from scratch when the server list changes */
+      pgmemcache_reset_context();
+      do_server_add(newval);
+    }
 }
 
 static void assign_default_behavior_guc(const char *newval, void *extra)
@@ -164,10 +189,8 @@ static void assign_default_behavior_guc(const char *newval, void *extra)
   StringInfoData flag_buf;
   StringInfoData data_buf;
   memcached_return rc;
-  MemoryContext old_ctx;
   if (!newval)
     return;
-  old_ctx = MemoryContextSwitchTo(globals.pg_ctxt);
 
   initStringInfo(&flag_buf);
   initStringInfo(&data_buf);
@@ -220,18 +243,6 @@ static void assign_default_behavior_guc(const char *newval, void *extra)
     }
   pfree(flag_buf.data);
   pfree(data_buf.data);
-
-  MemoryContextSwitchTo(old_ctx);
-}
-
-/* This is called when we're being unloaded from a process. Note that
- * this only happens when we're being replaced by a LOAD (e.g. it
- * doesn't happen on process exit), so we can't depend on it being
- * called. */
-void _PG_fini(void)
-{
-  memcached_free(globals.mc);
-  MemoryContextDelete(globals.pg_ctxt);
 }
 
 Datum memcache_add(PG_FUNCTION_ARGS)
@@ -325,25 +336,6 @@ Datum memcache_delete(PG_FUNCTION_ARGS)
   PG_RETURN_BOOL(rc == 0);
 }
 
-static time_t interval_to_time_t(Interval *span)
-{
-  float8 result;
-
-#ifdef HAVE_INT64_TIMESTAMP
-  result = span->time / 1000000e0;
-#else
-  result = span->time;
-#endif
-
-  if (span->month != 0)
-    {
-      result += (365.25 * 86400) * (span->month / 12);
-      result += (30.0 * 86400) * (span->month % 12);
-    }
-
-  return (time_t) result;
-}
-
 Datum memcache_flush_all0(PG_FUNCTION_ARGS)
 {
   static time_t opt_expire = 0;
@@ -390,6 +382,7 @@ Datum memcache_get(PG_FUNCTION_ARGS)
   ret = (text *) palloc(return_value_length + VARHDRSZ);
   SET_VARSIZE(ret, return_value_length + VARHDRSZ);
   memcpy(VARDATA(ret), string, return_value_length);
+  free(string);
 
   PG_RETURN_TEXT_P(ret);
 }
@@ -444,9 +437,9 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
       keys = palloc(sizeof(char *) * (array_length + 1 /* extra key for last memcached_fetch call */ ));
       key_lens = palloc(sizeof(size_t) * (array_length + 1));
 
-      // initialize terminating extra-key
-      keys[ array_length ] = 0;
-      key_lens[ array_length ] = 0;
+      /* initialize terminating extra-key */
+      keys[array_length] = 0;
+      key_lens[array_length] = 0;
 
       for (i = 0; i < array_length; i++)
         {
@@ -500,15 +493,17 @@ Datum memcache_get_multi(PG_FUNCTION_ARGS)
           SRF_RETURN_DONE(funcctx);
         }
       values = (char **) palloc(2 * sizeof(char *));
-      values[0] = (char *) palloc((fctx->key_lens[funcctx->call_cntr] + /* additional space for terminating zero character */ 1) * sizeof(char));
-      values[1] = (char *) palloc((value_length + /* additional space for terminating zero character */ 1) * sizeof(char));
+      /* make sure we have space for terminating zero character */
+      values[0] = (char *) palloc(fctx->key_lens[funcctx->call_cntr] + 1);
+      values[1] = (char *) palloc(value_length + 1);
 
       memcpy(values[0], fctx->keys[funcctx->call_cntr], fctx->key_lens[funcctx->call_cntr]);
       memcpy(values[1], value, value_length);
+      free(value);
 
-      // BuildTupleFromCStrings needs correct zero-terminated C-string, so terminate our raw strings
-      values[0][ fctx->key_lens[funcctx->call_cntr] ] = '\0';
-      values[1][ value_length ] = '\0';
+      /* BuildTupleFromCStrings needs correct zero-terminated C-string, so terminate our raw strings */
+      values[0][fctx->key_lens[funcctx->call_cntr]] = '\0';
+      values[1][value_length] = '\0';
 
       tuple = BuildTupleFromCStrings(attinmeta, values);
       result = HeapTupleGetDatum(tuple);
@@ -565,21 +560,20 @@ Datum memcache_append_absexpire(PG_FUNCTION_ARGS)
 
 static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
 {
-  text *key = NULL, *val;
-  size_t key_length, val_length;
-  time_t expire;
-  TimestampTz timestamptz;
-  struct pg_tm tm;
-  fsec_t fsec;
-  bool ret;
+  memcached_return rc = MEMCACHED_FAILURE;
+  const char *func = NULL;
+  char *key, *value;
+  text *key_text = NULL, *value_text;
+  size_t key_length, value_length;
+  time_t expiration = 0;
 
   if (PG_ARGISNULL(0))
     elog(ERROR, "pgmemcache: key cannot be NULL");
   if (PG_ARGISNULL(1))
     elog(ERROR, "pgmemcache: value cannot be NULL");
 
-  key = PG_GETARG_TEXT_P(0);
-  key_length = VARSIZE(key) - VARHDRSZ;
+  key_text = PG_GETARG_TEXT_P(0);
+  key_length = VARSIZE(key_text) - VARHDRSZ;
 
   /* These aren't really needed as we set libmemcached behavior to check for all invalid sets */
   if (key_length < 1)
@@ -587,19 +581,22 @@ static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
   if (key_length >= 250)
     elog(ERROR, "pgmemcache: key too long");
 
-  val = PG_GETARG_TEXT_P(1);
-  val_length = VARSIZE(val) - VARHDRSZ;
+  value_text = PG_GETARG_TEXT_P(1);
+  value_length = VARSIZE(value_text) - VARHDRSZ;
 
-  expire = (time_t) 0.0;
   if (PG_NARGS() >= 3 && PG_ARGISNULL(2) == false)
     {
       if (type & PG_MEMCACHE_TYPE_INTERVAL)
         {
           Interval *span = PG_GETARG_INTERVAL_P(2);
-          expire = interval_to_time_t(span);
+          expiration = interval_to_time_t(span);
         }
       else if (type & PG_MEMCACHE_TYPE_TIMESTAMP)
         {
+          TimestampTz timestamptz;
+          struct pg_tm tm;
+          fsec_t fsec;
+
           timestamptz = PG_GETARG_TIMESTAMPTZ(2);
 
           /* convert to timestamptz to produce consistent results */
@@ -609,9 +606,9 @@ static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
                      errmsg("timestamp out of range")));
 
 #ifdef HAVE_INT64_TIMESTAMP
-          expire = (time_t) ((timestamptz - SetEpochTimestamp()) / 1000000e0);
+          expiration = (time_t) ((timestamptz - SetEpochTimestamp()) / 1000000e0);
 #else
-          expire = (time_t) timestamptz - SetEpochTimestamp();
+          expiration = (time_t) timestamptz - SetEpochTimestamp();
 #endif
         }
       else
@@ -620,16 +617,8 @@ static Datum memcache_set_cmd(int type, PG_FUNCTION_ARGS)
         }
     }
 
-  ret = do_memcache_set_cmd(type, VARDATA(key), key_length, VARDATA(val), val_length, expire);
-
-  PG_RETURN_BOOL(ret);
-}
-
-static bool do_memcache_set_cmd(int type, char *key, size_t key_length,
-                                char *value, size_t value_length, time_t expiration)
-{
-  memcached_return rc = 1; /*FIXME GCC Warning hack*/
-  const char *func = NULL;
+  key = VARDATA(key_text);
+  value = VARDATA(value_text);
 
   if (type & PG_MEMCACHE_ADD)
     {
@@ -665,13 +654,14 @@ static bool do_memcache_set_cmd(int type, char *key, size_t key_length,
   if (rc != MEMCACHED_SUCCESS)
     elog(WARNING, "pgmemcache: %s: %s", func,
                   memcached_strerror(globals.mc, rc));
-  return rc == MEMCACHED_SUCCESS;
+
+  PG_RETURN_BOOL(rc == MEMCACHED_SUCCESS);
 }
 
 Datum memcache_server_add(PG_FUNCTION_ARGS)
 {
   text *server = PG_GETARG_TEXT_P(0);
-  char * host_str;
+  char *host_str;
   memcached_return rc;
 
   host_str = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(server)));
@@ -688,16 +678,11 @@ static memcached_return do_server_add(const char *host_str)
 {
   memcached_server_st *servers;
   memcached_return rc;
-  MemoryContext old_ctx;
-
-  old_ctx = MemoryContextSwitchTo(globals.pg_ctxt);
 
   servers = memcached_servers_parse(host_str);
   rc = memcached_server_push(globals.mc, servers);
   memcached_server_list_free(servers);
 
-  MemoryContextSwitchTo(old_ctx);
-
   return rc;
 }
 
@@ -712,6 +697,9 @@ static memcached_behavior get_memcached_behavior_flag(const char *value)
   MC_STR_TO_ENUM(BEHAVIOR, BUFFER_REQUESTS);
   MC_STR_TO_ENUM(BEHAVIOR, CACHE_LOOKUPS);
   MC_STR_TO_ENUM(BEHAVIOR, CONNECT_TIMEOUT);
+#if LIBMEMCACHED_VERSION_HEX >= 0x01000003
+  MC_STR_TO_ENUM(BEHAVIOR, DEAD_TIMEOUT);
+#endif
   MC_STR_TO_ENUM(BEHAVIOR, DISTRIBUTION);
   MC_STR_TO_ENUM(BEHAVIOR, HASH);
   MC_STR_TO_ENUM(BEHAVIOR, HASH_WITH_PREFIX_KEY);
@@ -727,6 +715,9 @@ static memcached_behavior get_memcached_behavior_flag(const char *value)
   MC_STR_TO_ENUM(BEHAVIOR, POLL_TIMEOUT);
   MC_STR_TO_ENUM(BEHAVIOR, RANDOMIZE_REPLICA_READ);
   MC_STR_TO_ENUM(BEHAVIOR, RCV_TIMEOUT);
+#if LIBMEMCACHED_VERSION_HEX >= 0x00049000
+  MC_STR_TO_ENUM(BEHAVIOR, REMOVE_FAILED_SERVERS);
+#endif
   MC_STR_TO_ENUM(BEHAVIOR, RETRY_TIMEOUT);
   MC_STR_TO_ENUM(BEHAVIOR, SERVER_FAILURE_LIMIT);
   MC_STR_TO_ENUM(BEHAVIOR, SND_TIMEOUT);
@@ -738,9 +729,6 @@ static memcached_behavior get_memcached_behavior_flag(const char *value)
   MC_STR_TO_ENUM(BEHAVIOR, USER_DATA);
   MC_STR_TO_ENUM(BEHAVIOR, USE_UDP);
   MC_STR_TO_ENUM(BEHAVIOR, VERIFY_KEY);
-#if LIBMEMCACHED_VERSION_HEX >= 0x00049000
-  MC_STR_TO_ENUM(BEHAVIOR, REMOVE_FAILED_SERVERS);
-#endif
 
   elog(ERROR, "pgmemcache: unknown behavior flag: %s", value);
   return -1;
@@ -818,10 +806,10 @@ static memcached_return_t server_stat_function(memcached_st *ptr,
     {
       char *value = memcached_stat_get_value(ptr, &stat, *stat_ptr, &rc);
       appendStringInfo(context, "%s: %s\n", *stat_ptr, value);
-      pfree(value);
+      free(value);
     }
 
-  pfree(list);
+  free(list);
   return MEMCACHED_SUCCESS;
 }
 
